@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2015-2016 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2015-2017 -- leonerd@leonerd.org.uk
 
 package Device::Chip::SSD1306;
 
@@ -9,19 +9,21 @@ use strict;
 use warnings;
 use base qw( Device::Chip );
 
-our $VERSION = '0.03';
+our $VERSION = '0.04';
+
+use Carp;
 
 =encoding UTF-8
 
 =head1 NAME
 
-C<Device::Chip::SSD1306> - chip driver for a F<SSD1306> OLED module
+C<Device::Chip::SSD1306> - chip driver for monochrome OLED modules
 
 =head1 DESCRIPTION
 
 This abstract L<Device::Chip> subclass provides communication to an F<SSD1306>
-chip attached by an adapter. To actually use it, you should use one of the
-subclasses for the various interface types.
+or F<SH1106> chip attached by an adapter. To actually use it, you should use
+one of the subclasses for the various interface types.
 
 =over 4
 
@@ -41,6 +43,91 @@ concepts or features, only the use of this module to access them.
 
 =cut
 
+=head1 DEVICE MODELS
+
+This module supports a variety of actual chip modules with different display
+sizes. The specific display module is selected by the C<model> argument to the
+constructor, which should take one of the following values:
+
+=over 4
+
+=item C<SSD1306-128x64>
+
+An F<SSD1306> driving a display with 128 columns and 64 rows. The most common
+of the display modules, often found with a 0.96 inch display.
+
+=item C<SSD1306-128x32>
+
+An F<SSD1306> driving a display with 128 columns and 32 rows. This is the
+usual "half-height" display.
+
+=item C<SH1106-128x64>
+
+An F<SH1106> driving a display with 128 columns and 64 rows. This is the chip
+that's usually found in the larger display modules, such as 1.3 and 1.6 inch.
+
+=back
+
+=cut
+
+my %MODELS = (
+   "SSD1306-128x64" => {
+      columns => 128,
+      rows    => 64,
+
+      set_com_pins_arg => 0x12,
+      column_offset    => 0,
+   },
+   "SSD1306-128x32" => {
+      columns => 128,
+      rows    => 32,
+
+      set_com_pins_arg => 0x02,
+      column_offset    => 0,
+   },
+   "SH1106-128x64" => {
+      columns => 128,
+      rows    => 64,
+
+      set_com_pins_arg => 0x12,
+      # SH1106 has a 128 column physical display but its VRAM is 132 columns
+      #   wide; with two blank columns either side of the display. It actually
+      #   starts outputting from column 2;
+      column_offset    => 2,
+   },
+);
+
+=head1 CONSTRUCTOR
+
+=cut
+
+=head2 new
+
+   $chip = Device::Chip::SSD1306->new(
+      model => $model,
+   )
+
+Returns a new C<Device::Chip::SSD1306> driver instance for the given model
+name, which must be one of the models listed in L</DEVICE MODELS>. If no model
+option is chosen, the default of C<SSD1306-128x64> will apply.
+
+=cut
+
+sub new
+{
+   my $class = shift;
+   my %args = @_;
+
+   my $self = $class->SUPER::new( %args );
+
+   my $modelargs = $MODELS{ $args{model} // "SSD1306-128x64" }
+      or croak "Unrecognised model $args{model}";
+
+   $self->{$_} = $modelargs->{$_} for keys %$modelargs;
+
+   return $self;
+}
+
 =head1 METHODS
 
 The following methods documented with a trailing call to C<< ->get >> return
@@ -48,6 +135,22 @@ L<Future> instances.
 
 =cut
 
+=head2 rows
+
+=head2 columns
+
+   $n = $chip->rows
+
+   $n = $chip->columns
+
+Simple accessors that return the number of rows or columns present on the
+physical display.
+
+=cut
+
+sub rows    { shift->{rows} }
+sub columns { shift->{columns} }
+
 use constant {
    CMD_SET_CONTRAST      => 0x81, #    , contrast
    CMD_DISPLAY_LAMPTEST  => 0xA4, # + all on
@@ -88,7 +191,9 @@ use constant {
    CMD_NOP             => 0xE3,
 };
 
-=head2 $ssd->init->get
+=head2 init
+
+   $chip->init->get
 
 Initialise the display after reset to some sensible defaults.
 
@@ -110,13 +215,15 @@ sub init
       ->then( sub { $self->send_cmd( CMD_SET_ADDR_MODE,    MODE_HORIZONTAL ) })
       ->then( sub { $self->send_cmd( CMD_SET_SEGMENT_REMAP | 1 ) })
       ->then( sub { $self->send_cmd( CMD_SET_COM_SCAN_DIR  | 0 ) })
-      ->then( sub { $self->send_cmd( CMD_SET_COM_PINS,     0 | 0x12 ) })
+      ->then( sub { $self->send_cmd( CMD_SET_COM_PINS,     $self->{set_com_pins_arg} ) })
       ->then( sub { $self->send_cmd( CMD_SET_CONTRAST,     0x9F ) })
       ->then( sub { $self->send_cmd( CMD_SET_PRECHARGE,    ( 0x0f << 4 ) | ( 1 ) ) })
       ->then( sub { $self->send_cmd( CMD_SET_VCOMH_LEVEL,  ( 4 << 4 ) ) });
 }
 
-=head2 $ssd->display( $on )->get
+=head2 display
+
+   $chip->display( $on )->get
 
 Turn on or off the display.
 
@@ -129,7 +236,9 @@ sub display
    $self->send_cmd( $on ? CMD_DISPLAY_ON : CMD_DISPLAY_OFF );
 }
 
-=head2 $ssd->display_lamptest( $enable )->get
+=head2 display_lamptest
+
+   $chip->display_lamptest( $enable )->get
 
 Turn on or off the all-pixels-lit lamptest mode.
 
@@ -143,10 +252,12 @@ sub display_lamptest
    $self->send_cmd( CMD_DISPLAY_LAMPTEST + !!$enable );
 }
 
-=head2 $ssd->send_display( $pixels )->get
+=head2 send_display
+
+   $chip->send_display( $pixels )->get
 
 Sends an entire screen-worth of pixel data. The C<$pixels> should be in a
-packed binary string 128 * 8 bytes long.
+packed binary string containing one byte per 8 pixels.
 
 =cut
 
@@ -155,9 +266,24 @@ sub send_display
    my $self = shift;
    my ( $pixels ) = @_;
 
-   $self->send_cmd( CMD_SET_COLUMN_ADDR, 0, 127 )
-      ->then( sub { $self->send_cmd( CMD_SET_PAGE_ADDR, 0, 7 ) })
-      ->then( sub { $self->send_data( $pixels ) })
+   # This output method isn't quite what most of the SSD1306 drivers use, but
+   # it happens to work on both the SSD1306 and the SH1106, whereas other code
+   # based on SET_COLUMN_ADDR + SET_PAGE_ADDR do not
+
+   my $f = Future->done;
+
+   my $pagewidth = $self->columns;
+
+   my $column = $self->{column_offset};
+
+   foreach my $page ( 0 .. ( $self->rows / 8 ) - 1 ) {
+      $f = $f->then( sub { $self->send_cmd( CMD_SET_PAGE_START + $page ) } )
+         ->then( sub { $self->send_cmd( CMD_SET_LOW_COLUMN | $column & 0x0f ) } )
+         ->then( sub { $self->send_cmd( CMD_SET_HIGH_COLUMN | $column >> 4 ) } )
+         ->then( sub { $self->send_data( substr $pixels, $page * $pagewidth, $pagewidth ) } );
+   }
+
+   return $f;
 }
 
 =head1 TODO
